﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.WpfDesktopEdition;

namespace DisplayIsoLinesWPF
{
    public partial class Sample : Window
    {
        private string wellDepthPointDataFilePath = @"..\..\Data\Test.csv";
        private Dictionary<PointShape, double> wellDepthPointData;

        //private GridCell[,] gridCellMatrix;
        private const string gridPath = @"..\..\Data\result.grid";

        private RectangleShape _defualtExtent;

        public Sample()
        {
            InitializeComponent();

            Map1.MapUnit = GeographyUnit.Meter;
        }

        private void LayoutRoot_Loaded(object sender, RoutedEventArgs e)
        {
            btnGenerateGridFile.IsEnabled = true;

            //Load the well depth points and depth data from a text file into the dictionary
            //We cache this at the class level to prevent form loading it multiple times
            wellDepthPointData = GetWellDepthPointDataFromCSV(wellDepthPointDataFilePath, 10000000);
            txtSearch.Text = (wellDepthPointData.Select(w => w.Value).Max() + 1).ToString();

            ShapeFileFeatureLayer shape = new ShapeFileFeatureLayer(@"..\..\Data\Cardiff.shp");
            shape.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle = new AreaStyle(new GeoPen(GeoColor.FromArgb(255, new GeoColor(0, 0, 0)), 5), new GeoSolidBrush(new GeoColor(255, 255, 255)));
            shape.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;

            shape.Open();
            _defualtExtent = shape.GetBoundingBox();
            Map1.CurrentExtent = _defualtExtent;
            shape.Close();

            //Add the grid layer, the grid cells, and the well points to the map
            LayerOverlay isoLineOverlay = new LayerOverlay();

            Map1.Overlays.Add("isoLineOverlay", isoLineOverlay);

            isoLineOverlay.Layers.Add("shp", shape);

            isoLineOverlay.Layers.Add("IsoLineLayer", GetGridIsoLineLayer());
            isoLineOverlay.Layers.Add("GridCellsLayer", GetGridFeatureLayer());

            //  isoLineOverlay.Layers.Add("WellsLayer", GetWellDepthPointLayer());

            isoLineOverlay.TileType = TileType.MultipleTile;
            Map1.Refresh();
        }

        private static Dictionary<PointShape, double> GetWellDepthPointDataFromCSV(string csvFilePath, double noDataValue)
        {
            StreamReader streamReader = null;
            Dictionary<PointShape, double> points = new Dictionary<PointShape, double>();

            try
            {
                streamReader = new StreamReader(csvFilePath);

                while (!streamReader.EndOfStream)
                {
                    string line = streamReader.ReadLine();

                    string[] parts = line.Split(',');
                    points.Add(new PointShape(double.Parse(parts[0]), double.Parse(parts[1])), double.Parse(parts[2]));
                }
            }
            finally
            {
                if (streamReader != null) { streamReader.Close(); }
            }

            return points.Where(p => p.Value != noDataValue).ToDictionary(o => o.Key, o => o.Value);
        }

        private void CreateGridCellMatrix(RectangleShape extent)
        {
            //Get the current extent since we use that to gerenate the grid.  Of course this is just for
            //the demo and in the real world you can use any extent you like
            // RectangleShape currentDrawingExtent = ExtentHelper.GetDrawingExtent(Map1.CurrentExtent, (float)Map1.ActualWidth, (float)Map1.ActualHeight);

            //Calculate the cell size based on how many rows and columns you specified
            Stopwatch watch = Stopwatch.StartNew();
            double cellSize = Math.Min(extent.Width / double.Parse(txtGridIsoLineCellColumnCount.Text), extent.Height / Int32.Parse(txtGridIsoLineCellRowCount.Text));

            //Greate the grid definition based on the extent, cell size etc.
            GridDefinition gridDefinition = new GridDefinition(extent, cellSize, -9999, wellDepthPointData);

            //Generate the grid based on Inverse Distance Weighted interpolation model.  You can define your own model if needed.

            txtSearch.Text = (cellSize).ToString();
            double searchRadious = double.Parse(txtSearch.Text);

            FileStream writer = null;
            try
            {
                writer = new FileStream(gridPath, FileMode.Create);
                GridFeatureSource.GenerateGrid(gridDefinition, new InverseDistanceWeightedGridInterpolationModel(double.Parse(txtPower.Text), searchRadious), writer);
            }
            finally
            {
                if (writer != null) writer.Close();
            }

            watch.Stop();

            //timeListview.Items.Add(String.Format("R:{0}-L:{1}-P:{2}-S:{3}-T:{4}", txtGridIsoLineCellRowCount.Text,
            //                                     txtGridIsoLineLevelCount.Text, txtPower.Text, txtSearch.Text, watch.Elapsed));
            timeListview.Items.Add(String.Format("Time:{0}", watch.Elapsed));
        }

        private GridIsoLineLayer GetGridIsoLineLayer()
        {
            //Create a grid cell matrix based on the textboxes and the current extent
            CreateGridCellMatrix(_defualtExtent);

            //Load the grid we just created into the GridIsoLineLayer using the number of breaks defines
            //on the screen

            Collection<double> isoLineLevels = IsoLineLayer.GetIsoLineLevels(wellDepthPointData, Convert.ToInt32(txtGridIsoLineLevelCount.Text));

            GridIsoLineLayer isoLineLayer = new GridIsoLineLayer(gridPath, isoLineLevels);

            //Create a series of colors from blue to red that we will use for the breaks
            Collection<GeoColor> colors = GeoColor.GetColorsInQualityFamily(GeoColor.StandardColors.Blue, GeoColor.StandardColors.Red, isoLineLevels.Count, ColorWheelDirection.Clockwise);

            //Setup a class break style based on the isoline levels and the colors
            ClassBreakStyle classBreakLineStyle = new ClassBreakStyle(isoLineLayer.DataValueColumnName);
            LineStyle firstStyle = new LineStyle(new GeoPen(colors[0], 3));
            classBreakLineStyle.ClassBreaks.Add(new ClassBreak(double.MinValue, firstStyle));
            for (int i = 0; i < colors.Count - 1; i++)
            {
                LineStyle style = new LineStyle(new GeoPen(colors[i + 1], 3));
                classBreakLineStyle.ClassBreaks.Add(new ClassBreak(isoLineLevels[i], style));
            }
            isoLineLayer.CustomStyles.Add(classBreakLineStyle);

            //Create the text styles to label the lines
            //TextStyle textStyle = TextStyles.CreateSimpleTextStyle(isoLineLayer.DataValueColumnName, "Arial", 5, DrawingFontStyles.Bold, GeoColor.StandardColors.Black, 0, 0);
            //textStyle.HaloPen = new GeoPen(GeoColor.StandardColors.White, 2);
            //textStyle.OverlappingRule = LabelOverlappingRule.NoOverlapping;
            //textStyle.SplineType = SplineType.StandardSplining;
            //textStyle.DuplicateRule = LabelDuplicateRule.UnlimitedDuplicateLabels;
            //textStyle.TextLineSegmentRatio = 9999999;
            //textStyle.FittingLineInScreen = true;
            //textStyle.SuppressPartialLabels = true;
            //isoLineLayer.CustomStyles.Add(textStyle);

            return isoLineLayer;
        }

        private GridFeatureLayer GetGridFeatureLayer()
        {
            //Get the line breaks
            Collection<double> isoLineBreaks = IsoLineLayer.GetIsoLineLevels(wellDepthPointData, Convert.ToInt32(txtGridIsoLineLevelCount.Text));

            //Load a new GridFeatureLayer based on the current grid file
            GridFeatureLayer gridFeatureLayer = new GridFeatureLayer(gridPath);

            //Create a series of colours from blue to red that we will use for the breaks
            Collection<GeoColor> colors = GeoColor.GetColorsInQualityFamily(GeoColor.StandardColors.Blue, GeoColor.StandardColors.Red, isoLineBreaks.Count, ColorWheelDirection.Clockwise);

            //Create a class break style
            ClassBreakStyle classBreakLineStyle = new ClassBreakStyle(gridFeatureLayer.DataValueColumnName);

            //Setup a class break style based on the isoline levels and the colours
            AreaStyle firstStyle = new AreaStyle(new GeoPen(GeoColor.FromArgb(50, colors[0]), 1), new GeoSolidBrush(GeoColor.FromArgb(50, colors[0])));
            classBreakLineStyle.ClassBreaks.Add(new ClassBreak(double.MinValue, firstStyle));
            for (int i = 1; i < colors.Count - 1; i++)
            {
                AreaStyle style = new AreaStyle(new GeoPen(GeoColor.FromArgb(50, colors[i]), 1), new GeoSolidBrush(GeoColor.FromArgb(50, colors[i])));
                classBreakLineStyle.ClassBreaks.Add(new ClassBreak(isoLineBreaks[i], style));
            }

            gridFeatureLayer.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(classBreakLineStyle);
            gridFeatureLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            return gridFeatureLayer;
        }

        private InMemoryFeatureLayer GetWellDepthPointLayer()
        {
            //Create an in memory layer to hold the well data from the text file
            InMemoryFeatureLayer inMemoryFeatureLayer = new InMemoryFeatureLayer();
            inMemoryFeatureLayer.FeatureSource.Open();

            //Make sure to specify the depth column
            inMemoryFeatureLayer.Columns.Add(new FeatureSourceColumn("Depth", "String", 10));

            //Loop through all the point data and add it to the in memory layer
            foreach (KeyValuePair<PointShape, double> wellDepth in wellDepthPointData)
            {
                Feature feature = new Feature(wellDepth.Key);
                feature.ColumnValues.Add("Depth", wellDepth.Value.ToString());
                inMemoryFeatureLayer.InternalFeatures.Add(feature);
            }

            //Now that all of the data is added we can build an in memory index to make the lookups fast
            inMemoryFeatureLayer.BuildIndex();

            //Create the well point style
            PointStyle pointStyle1 = PointStyles.CreateSimpleCircleStyle(GeoColor.StandardColors.White, 1, GeoColor.SimpleColors.Black, 1);
            inMemoryFeatureLayer.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(pointStyle1);

            //Create the text style with a halo
            TextStyle textStyle = TextStyles.CreateSimpleTextStyle("Depth", "Arial", 5, DrawingFontStyles.Regular, GeoColor.SimpleColors.Black);
            textStyle.HaloPen = new GeoPen(GeoColor.StandardColors.White, 3);
            textStyle.PointPlacement = PointPlacement.UpperCenter;
            textStyle.YOffsetInPixel = 5;

            //Apply these styles at all levels and add then to the custom styles for the layer
            inMemoryFeatureLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            inMemoryFeatureLayer.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(textStyle);

            return inMemoryFeatureLayer;
        }

        #region Control Events

        private void btnGenerateGridFile_Click(object sender, RoutedEventArgs e)
        {
            btnGenerateGridFile.IsEnabled = true;

            //Close the previous isoLineLayer and get ready to replace it with a new one
            LayerOverlay layerOverlay = (LayerOverlay)Map1.Overlays["isoLineOverlay"];
            layerOverlay.Layers["IsoLineLayer"].Close();
            layerOverlay.Layers["GridCellsLayer"].Close();

            layerOverlay.Layers["IsoLineLayer"] = GetGridIsoLineLayer();
            layerOverlay.Layers["GridCellsLayer"] = GetGridFeatureLayer();
            layerOverlay.Layers["GridCellsLayer"].IsVisible = cbxGridLine.IsChecked.Value;

            cbxGridLine.IsEnabled = true;
            layerOverlay.Refresh();
        }

        private void cbxGridLine_Checked(object sender, RoutedEventArgs e)
        {
            if (Map1.Overlays.Contains("isoLineOverlay"))
            {
                LayerOverlay layerOverlay = (LayerOverlay)Map1.Overlays["isoLineOverlay"];
                layerOverlay.Layers["GridCellsLayer"].IsVisible = true;
                layerOverlay.Refresh();
            }
        }

        private void cbxGridLine_Unchecked(object sender, RoutedEventArgs e)
        {
            LayerOverlay layerOverlay = (LayerOverlay)Map1.Overlays["isoLineOverlay"];
            layerOverlay.Layers["GridCellsLayer"].IsVisible = false;
            layerOverlay.Refresh();
        }

        #endregion Control Events
    }
}